/**
  *  \file planetscentral/dbclient/dbclient.cc
  *  \brief Database Client
  */

#include <stdexcept>
#include <cstring>
#include <cstdio>
#include <unistd.h>
#include "planetscentral/dbclient/dbclient.h"
#include "util/misc.h"

/******************************* DbRequest *******************************/

DbRequest&
DbRequest::withString(const string_t& str)
{
    pushNew(makeStringValue(str));
    return *this;
}

DbRequest&
DbRequest::withString(const char* str)
{
    pushNew(makeStringValue(str));
    return *this;
}

DbRequest&
DbRequest::withInt(int32_t n)
{
    pushNew(makeIntValue(n));
    return *this;
}

DbRequest&
DbRequest::withInlineStringList(const StringList_t& sl)
{
    for (StringList_t::const_iterator i = sl.begin(); i != sl.end(); ++i) {
        pushNew(makeStringValue(*i));
    }
    return *this;
}

DbRequest&
DbRequest::withInlineIntList(const IntList_t& il)
{
    for (IntList_t::const_iterator i = il.begin(); i != il.end(); ++i) {
        pushNew(makeIntValue(*i));
    }
    return *this;
}

DbRequest&
DbRequest::withRawValue(IntValue* val)
{
    push(val);
    return *this;
}

/** Call database.
    \param cmd Command.
    \return newly-allocated result object */
IntValue*
DbClient::call(const IntDataSegment& cmd)
{
    checkConnected();

    // Send request
    sendArray(cmd, cmd.size());
    flush();

    // Read reply
    return readObject(false);
}

/** Call database, interpret result as integer. */
int32_t
DbClient::callInt(const IntDataSegment& cmd)
{
    std::auto_ptr<IntValue> v(call(cmd));
    return toIntResult(v.get());
}

/** Call database, interpret result as string. */
string_t
DbClient::callString(const IntDataSegment& cmd)
{
    std::auto_ptr<IntValue> v(call(cmd));
    return IntValue::toStringOf(v.get(), false);
}

/** Call database, discard result. */
void
DbClient::callVoid(const IntDataSegment& cmd)
{
    delete call(cmd);
}

/** Probe database readiness. */
void
DbReadyClient::onConnect()
{
    int count = 0;
    while (1) {
        try {
            callVoid(DbRequest().withString("PING"));
            break;
        }
        catch (std::exception& e) {
            if (std::strncmp(e.what(), "LOADING", 7) == 0) {
                // 10 x 1 second
                // 10 x 5 seconds
                // 10 x 20 seconds
                if (count > 30) {
                    // 260 seconds should be enough
                    std::printf("(not ready yet, giving up)\n");
                    throw;
                }
            } else {
                throw;
            }
        }
        int sleepTime = count > 20 ? 20 : count > 10 ? 5 : 1;
        std::printf("(not ready yet, sleeping %d seconds)\n", sleepTime);
        sleep(sleepTime);
    }
}
