/**
  *  \file planetscentral/talk/cmdgroup.cc
  *  \brief Talk: GROUP commands
  */

#include "planetscentral/talk/cmdgroup.h"
#include "int/args.h"
#include "int/array.h"
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/talk/group.h"
#include "planetscentral/talk/talk.h"
#include "planetscentral/talk/talkconnection.h"
#include "planetscentral/talk/db.h"

static void
configureGroup(Group& group, IntArgBlock& args)
{
    while (args.getNumArgs() > 0) {
        args.checkArgumentCountAtLeast(2);
        string_t key(toStringResult(args.getNext()));
        string_t value(toStringResult(args.getNext()));
        if (key == "parent") {
            group.setParent(value);
        } else {
            group.header().setField(key, value);
        }
    }
}

void
doGroupAdd(TalkConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCountAtLeast(3);
    string_t groupId = toStringResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    Group group(database_connection, groupId);
    conn.checkAdmin();
    if (group.exists()) {
        conn.sendError("409 Already exists");
    } else {
        configureGroup(group, args);
        conn.sendSuccess("OK");
    }
}

void
doGroupSet(TalkConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCountAtLeast(1);
    string_t groupId = toStringResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    Group group(database_connection, groupId);
    conn.checkAdmin();
    if (!group.exists()) {
        conn.sendError("404 Not found");
    } else {
        configureGroup(group, args);
        conn.sendSuccess("OK");
    }
}

void
doGroupGet(TalkConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCount(2);
    string_t groupId = toStringResult(args.getNext());
    string_t field = toStringResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    std::auto_ptr<IntValue> result(Group(database_connection, groupId).header().getFieldRaw(field));

    // Send result
    conn.sendObject(result.get());
}

void
doGroupList(TalkConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCount(1);
    string_t groupId = toStringResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    Group group(database_connection, groupId);
    Ptr<IntArrayData> ad(new IntArrayData());

    if (!conn.isAdmin() && group.unlisted().get()) {
        Ptr<IntArrayData> blank(new IntArrayData());
        blank->addDimension(0);
        ad->content.pushNew(makeStringValue("groups"));
        ad->content.pushNew(new IntArray(blank));
        ad->content.pushNew(makeStringValue("forums"));
        ad->content.pushNew(new IntArray(blank));
    } else {
        ad->content.pushNew(makeStringValue("groups"));
        ad->content.pushNew(group.subgroups()
                            .sort()
                            .sortLexicographical()
                            .by(Group(database_connection, "*").key())
                            .getResult());
        ad->content.pushNew(makeStringValue("forums"));
        ad->content.pushNew(group.forums()
                            .sort()
                            .sortLexicographical()
                            .by(FORUM_ROOT "*:header->key")
                            .getResult());
    }
    ad->addDimension(ad->content.size());

    // Send result
    IntArray arr(ad);
    conn.sendObject(&arr);
}

void
doGroupStat(TalkConnection& conn, IntArgBlock& args)
{
    // Parse args
    args.checkArgumentCount(1);
    string_t groupId = toStringResult(args.getNext());

    // Do it
    ReconnectingGuard g(database_rc);
    Group group(database_connection, groupId);
    std::auto_ptr<IntValue> result(group.describe(conn.getRenderState()));

    // Send result
    conn.sendObject(result.get());
}

void
doGroupMStat(TalkConnection& conn, IntArgBlock& args)
{
    // Do it
    ReconnectingGuard g(database_rc);
    Ptr<IntArrayData> ad(new IntArrayData());
    while (args.getNumArgs() > 0) {
        string_t groupId = toStringResult(args.getNext());
        ad->content.pushNew(Group(database_connection, groupId).describe(conn.getRenderState()));
    }
    ad->addDimension(ad->content.size());

    // Send result
    IntArray arr(ad);
    conn.sendObject(&arr);
}
