/**
  *  \file planetscentral/talk/message.cc
  */

#include <time.h>
#include "planetscentral/talk/message.h"
#include "planetscentral/talk/db.h"
#include "planetscentral/talk/talk.h"
#include "cpluslib/format.h"
#include "planetscentral/talk/topic.h"
#include "int/array.h"
#include "cpluslib/misc.h"
#include "planetscentral/talk/user.h"
#include "planetscentral/talk/forum.h"
#include "util/misc.h"
#include "planetscentral/net/encoding.h"

static string_t wrapAngles(string_t id)
{
    return "<" + id + ">";
}


Message::Message(DbClient& client, int32_t messageId)
    : message(DbSubtree(client, MESSAGE_ROOT).subtree(messageId)),
      messageId(messageId)
{ }

DbHash
Message::header()
{
    return message.hashKey("header");
}

DbHashIntField
Message::topicId()
{
    return header().intField("thread");
}

DbHashIntField
Message::parentMessageId()
{
    return header().intField("parent");
}

DbHashIntField
Message::postTime()
{
    return header().intField("time");
}

DbHashIntField
Message::editTime()
{
    return header().intField("edittime");
}

DbHashStringField
Message::author()
{
    return header().stringField("author");
}

DbHashStringField
Message::subject()
{
    return header().stringField("subject");
}

DbHashStringField
Message::rfcMessageId()
{
    return header().stringField("msgid");
}

DbHashStringField
Message::rfcHeaders()
{
    return header().stringField("rfcheader");
}

DbHashIntField
Message::sequenceNumber()
{
    return header().intField("seq");
}

DbHashIntField
Message::previousSequenceNumber()
{
    return header().intField("prevseq");
}

DbHashStringField
Message::previousRfcMessageId()
{
    return header().stringField("prevmsgid");
}

DbString
Message::text()
{
    return message.stringKey("text");
}

bool
Message::exists()
{
    return header().exists();
}

void
Message::remove()
{
    // Remove from sets, so it becomes invisible
    Topic t(topic());
    Forum f(t.forum());
    t.messages().remove(messageId);
    f.messages().remove(messageId);
    User(message.getClient(), author().get()).postedMessages().remove(messageId);

    // Remove from NNTP side
    removeRfcMessageId(message.getClient(), rfcMessageId().get());

    // If the topic is now empty, remove it completely
    if (t.messages().empty()) {
        t.removeEmpty();
    }

    // Remove post
    text().remove();
    header().remove();
}

Topic
Message::topic()
{
    return Topic(message.getClient(), topicId().get());
}

int32_t
Message::getId() const
{
    return messageId;
}

IntValue*
Message::describe()
{
    // FIXME: can we use HMGET?
    const char*const fields[] = { "thread", "parent", "time", "edittime", "author", "subject" };
    Ptr<IntArrayData> ad = new IntArrayData();
    for (std::size_t i = 0; i < countof(fields); ++i) {
        ad->content.pushNew(makeStringValue(fields[i]));
        ad->content.pushNew(header().getFieldRaw(fields[i]));
    }
    ad->content.pushNew(makeStringValue("msgid"));
    ad->content.pushNew(makeStringValue(getRfcMessageId()));
    ad->addDimension(ad->content.size());
    return new IntArray(ad);
}

string_t
Message::getRfcMessageId()
{
    string_t msgid = rfcMessageId().get();
    if (msgid.empty()) {
        msgid = format("%d.%d%s", messageId, sequenceNumber().get(), msgid_suffix);
    }
    return msgid;
}

string_t
Message::getPreviousRfcMessageId()
{
    string_t msgid = previousRfcMessageId().get();
    if (msgid.empty()) {
        msgid = format("%d.%d%s", messageId, previousSequenceNumber().get(), msgid_suffix);
    }
    return msgid;
}

IntValue*
Message::getRfcHeader()
{
    Topic t(topic());
    Forum f(t.forum());
    string_t userId(author().get());
    User u(message.getClient(), userId);

    Ptr<IntArrayData> head(new IntArrayData());

    // Id pseudo-header
    head->content.pushNew(makeStringValue(":Id"));
    head->content.pushNew(makeIntValue(messageId));

    // Sequence pseudo-header
    int32_t seq = sequenceNumber().get();
    head->content.pushNew(makeStringValue(":Seq"));
    head->content.pushNew(makeIntValue(seq));

    // Xref
    head->content.pushNew(makeStringValue("Xref"));
    head->content.pushNew(makeStringValue(format("%s %s:%d", path_host, f.getNewsgroup(), seq)));

    // Path
    head->content.pushNew(makeStringValue("Path"));
    head->content.pushNew(makeStringValue(format("%s!not-for-mail", path_host)));

    // Message Id
    head->content.pushNew(makeStringValue("Message-Id"));
    head->content.pushNew(makeStringValue(wrapAngles(getRfcMessageId())));

    // From
    string_t userName(u.getLoginName());
    string_t email;
    if (u.profile().getIntField("infoemailflag")) {
        email = u.profile().getStringField("email");
        if (!email.empty()) {
            if (DbHash(message.getClient(), format("email:%s:status", email)).getStringField(format("status/%s", userId)) != "c") {
                email.clear();
            }
        }
    }
    if (email.empty()) {
        email = userName + "@invalid.invalid";
    }

    string_t realName = u.getRealName();
    if (realName.empty()) {
        realName = u.getScreenName();
    }
    head->content.pushNew(makeStringValue("From"));
    head->content.pushNew(makeStringValue(format("%s <%s>",
                                                 encodeMimeHeader(realName, "UTF-8"),
                                                 encodeMimeHeader(email, "UTF-8"))));

    // Newsgroups
    head->content.pushNew(makeStringValue("Newsgroups"));
    head->content.pushNew(makeStringValue(f.getNewsgroup()));

    // Subject
    head->content.pushNew(makeStringValue("Subject"));
    head->content.pushNew(makeStringValue(encodeMimeHeader(subject().get(), "UTF-8")));

    // Date
    int32_t pt = postTime().get();
    int32_t et = editTime().get();
    time_t efft = (et ? et : pt) * 60;       /* FIXME: hard-coded *60, xref getTime */
    struct tm efftm;
    char buf[100];
    gmtime_r(&efft, &efftm);
    strftime(buf, sizeof(buf), "%a, %d %b %Y %H:%M:%S +0000", &efftm);
    head->content.pushNew(makeStringValue("Date"));
    head->content.pushNew(makeStringValue(buf));

    // References
    int32_t parent = parentMessageId().get();
    if (parent != 0) {
        string_t refs;

        // Fetch 5 message Ids
        for (int i = 0; i < 5 && parent != 0; ++i) {
            Message m(message.getClient(), parent);
            if (!refs.empty()) {
                refs.insert(0, "\r\n ");
            }
            refs.insert(0, wrapAngles(m.getRfcMessageId()));
            parent = m.parentMessageId().get();
        }

        // Still more to do? Get thread starter
        if (parent != 0) {
            Message m(message.getClient(), t.firstPostingId().get());
            if (!refs.empty()) {
                refs.insert(0, "\r\n ");
            }
            refs.insert(0, wrapAngles(m.getRfcMessageId()));
        }
        head->content.pushNew(makeStringValue("References"));
        head->content.pushNew(makeStringValue(refs));
    }

    // Fake a bytes/lines size. For a precise value, we'd have to render the posting.
    // But let's keep it simple, it's used for a preview whether this is a long posting
    // or not, anyway, and I hope nobody uses it for allocating buffers.
    int32_t bytes = text().get().size();
    head->content.pushNew(makeStringValue(":Bytes"));
    head->content.pushNew(makeIntValue(bytes));
    head->content.pushNew(makeStringValue(":Lines"));
    head->content.pushNew(makeIntValue(bytes/40 + 1));

    // MIME
    head->content.pushNew(makeStringValue("MIME-Version"));
    head->content.pushNew(makeStringValue("1.0"));

    head->content.pushNew(makeStringValue("Content-Type"));
    head->content.pushNew(makeStringValue("text/plain; charset=UTF-8"));

    head->content.pushNew(makeStringValue("Content-Transfer-Encoding"));
    head->content.pushNew(makeStringValue("quoted-printable"));

    // Extras
    head->content.pushNew(makeStringValue("X-PCC-User"));
    head->content.pushNew(makeStringValue(userName));

    head->content.pushNew(makeStringValue("X-PCC-Posting-Id"));
    head->content.pushNew(makeIntValue(messageId));

    head->addDimension(head->content.size());
    return new IntArray(head);
}

int32_t
Message::allocateMessage(DbClient& client)
{
    return ++DbSubtree(client, MESSAGE_ROOT).intKey("id");
}

int32_t
Message::getLastId(DbClient& client)
{
    return DbSubtree(client, MESSAGE_ROOT).intKey("id").get();
}

/** Remove the link from a RFC message Id to a posting. */
void
Message::removeRfcMessageId(DbClient& client, string_t id)
{
    if (!id.empty()) {
        DbSubtree(client, RFC_MSGID_ROOT).intKey(id).remove();
    }
}

/** Add a link from a RFC message Id to a posting. */
void
Message::addRfcMessageId(DbClient& client, string_t id, int32_t messageId)
{
    if (!id.empty() && messageId != 0) {
        DbSubtree(client, RFC_MSGID_ROOT).intKey(id).set(messageId);
    }
}

/** Look up a RFC message Id. */
int32_t
Message::lookupRfcMessageId(DbClient& client, string_t rfcMsgId)
{
    string_t::size_type dot = rfcMsgId.find('.');
    int32_t result;

    if (rfcMsgId.empty()) {
        // error
        result = 0;
    } else if (dot != string_t::npos
        && rfcMsgId.size() - dot > msgid_suffix.size()
        && rfcMsgId.compare(rfcMsgId.size() - msgid_suffix.size(), msgid_suffix.size(), msgid_suffix) == 0
        && stringValue(rfcMsgId.substr(0, dot), result)
        && result > 0)
    {
        // a message Id generated by us
    } else {
        // external Id
        result = DbSubtree(client, RFC_MSGID_ROOT).intKey(rfcMsgId).get();
    }

    // Cross-check
    if (result > 0 && Message(client, result).getRfcMessageId() == rfcMsgId) {
        return result;
    } else {
        return 0;
    }
}
